/*
 * Copyright 2005-2006 Myriad Genetics, Inc.
 * Copyright 2008 Dustin Cartwright
 *
 * All rights reserved.
 */

#include "Phasing.h"
#include "genmap.h"
#include "twoPoint.h"
#include <string.h>
#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <float.h>

#define INITIAL_LOD_THRESH 3.0
#define BIG_INCREMENT 0.8
#define NUM_REFINEMENTS 3

#define MAX_NUM_FRONTIERS 18

#define MAX_NUM_NODE_PAIRS 2
#define MAX_NUM_NODES (2 * MAX_NUM_NODE_PAIRS)
#define MAX_NUM_LODS (1 << MAX_NUM_NODE_PAIRS)

enum {
    Obs0000 = 0,
    Obs0001 = 1,
    Obs0010 = 2,
    Obs0011 = 3,
    Obs0100 = 4,
    Obs0101 = 5,
    Obs0110 = 6,
    Obs0111 = 7,
    Obs1000 = 0x8,
    Obs1001 = 0x9,
    Obs1010 = 0xa,
    Obs1011 = 0xb,
    Obs1100 = 0xc,
    Obs1101 = 0xd,
    Obs1110 = 0xe,
    Obs1111 = 0xf
};

/* This describes an arbitrary function from a set of relative phases to lod
   scores. The array nodeNdxs, of size 2 * numNodePairs has all the node pairs.
   The array lods, of size 2 ** numNodePairs has is indexed by bit fields which
   correspond to the xor of the phases at the corresponding node pair. */
typedef struct {
    int     numNodePairs;
    int     nodeNdxs[MAX_NUM_NODES];
    double  lods[MAX_NUM_LODS];
         /* The difference between the largest and smallest LOD scores. */
    double  magnitude;
} relation;

typedef struct {
    int     fatherNode;
    int     motherNode;
} markerId;

struct linkageRelations_st {
    int         numNodes;
        /* 0-based array of size ped->numMarks which maps from marker ids to
           node ids. */
    markerId    *markerIds;
    int         numRelations;
        /* 0-based array of size numRelations */
    relation    *relations;
};
typedef struct linkageRelations_st linkageRelations;

typedef unsigned char *solution_t;

/* represents a partial solution to the phasing problem */
typedef struct {
    solution_t  solution;
    double      lod;
} partial_t;

/* This represents a family of partial solutions to the maximum likelihood
   problem. The array partials is of length 2 ** numFrontiers. The index
   is a bit field of length numFrontiers, which correspond to the phases of
   the nodes in the frontierNodes. Given the phases indicated by the index,
   the partial solution at that index is optimal with the given lod score.
 */
typedef struct {
    int         numFrontiers;
    int         *frontierNodes;
    partial_t   *partials;
    int         numNodes;
} partialsState_t;

/* Used by finalPhasing to represent a link between two groups of phases. */
typedef struct {
    int     ndx1;
    int     ndx2;
    double  strength;
    bool    parity;
} link_t;

/* Used by finalPhasing to represent the relative phase between two groups of
   phases. */
typedef struct {
    int     group;
    bool    parity;
} relPhase_t;

bool phasingFromThreshRelations(pedigree_t *ped,
                                const linkageRelations *relations,
                                double thresh, solution_t *pSwaps);
bool phasingFromRelations(pedigree_t *ped, const linkageRelations *relations,
                          solution_t *pSwaps);

linkageRelations *buildLinkageRelations(pedigree_t *, double lodThresh,
                                        swap_t *resultSwaps);
bool buildRelation(pedigree_t *, int m1, int m2, bool motherPhase,
                   bool fatherPhase, relation *rel, const markerId *markIds,
                   double lodThresh);
void shiftRelations(linkageRelations *rel);
linkageRelations *threshRelations(const linkageRelations *src, double thresh);
void freeRelations(linkageRelations *relations);

int *findGroups(const linkageRelations *rels);
int *findDirectGroups(const linkageRelations *rels);
int findRoot(int *groups, int ndx);

int *findNodeOrder(const linkageRelations *);
double *computeNodeTotalMagnitudes(const linkageRelations *rels);
int **computeRelsNdx(const linkageRelations *rels, int **pRelsNdxCounts);
int *findNodeOrderFrom(const double *nodeTotalMangitudes,
                       const int *const *relsNdx, const int *relsNdxCounts,
                       int start, int *pMaxFrontier, int *numNodesGrp,
                       const linkageRelations *);
int pickNextNode(const linkageRelations *, const bool *used,
                 const int *numFrontierCrosses, const double *nodeMagnitudes,
                 const double *nodeMagnitudes2);
void updateNumFrontierCrosses(const bool *used, int *pNumFrontierCrosses,
                              int *numFrontier, const int *const *relsNdx,
                              const int *relsNdxCounts,
                              const linkageRelations *rels, int node);
void updateNodeMagnitudes(int *numUnknowns, double *nodeMagnitudes,
                          const linkageRelations *rels, const int *counts,
                          const int *const *relsNdx, int node);

solution_t maxLlhdPhases(const linkageRelations *, const int *nodeOrder,
                         int maxNumFrontier);
bool nodeIsArbitrary(const int *groups, int numNodes, const bool *used,
                     int node);
bool isBetweenGroups(const bool *used, const bool *applied,
                     const linkageRelations *rels);
void doublePartials(partialsState_t *st, int nodeNdx);
void applyRelation(const relation *, partialsState_t *st);
double relationValue(const relation *, solution_t);
void collapseNode(partialsState_t *st, int node);
void prematureCollapse(partialsState_t *st, double *pThresh, bool force);

void finalPhasing(const linkageRelations *rels, pedigree_t *ped,
                  swap_t *resultSwaps);
void linkGroups(link_t *links, int count, int i, int j, double strength,
                bool parity);
int compareLinks(const void *p1, const void *p2);
int findRootRelPhase(relPhase_t *groups, int ndx, bool *pParity);

void swapPhases(pedigree_t *, solution_t swaps,
                const linkageRelations *relations, swap_t *resultSwaps);
int countSwaps(solution_t swaps, int numNodes);

solution_t newSolution(int numNodes);
void freeSolution(solution_t);
solution_t copySolution(solution_t, int numNodes);
void setSolution(solution_t, int pos, bool val);
bool testSolution(solution_t, int pos);

/* Attempts to find optimal phases for the markers in pedigree ped.
 */
bool phasing(pedigree_t *ped, double lodThresh, swap_t *resultSwaps)
{
    linkageRelations    *relations;
    solution_t          swaps = NULL;
    bool                success = false;
    
    relations = buildLinkageRelations(ped, lodThresh, resultSwaps);
    
    if (phasingFromRelations(ped, relations, &swaps)) {
        swapPhases(ped, swaps, relations, resultSwaps);
        freeSolution(swaps);
        finalPhasing(relations, ped, resultSwaps);
        success = true;
    }
        
    delete [] relations->markerIds;
    delete [] relations->relations;
    delete relations;
    return success;
}

/* Attempts to find optimal phases for the markers in pedigree ped. It begins
   by using a threshold 
 */
double phasingVary(pedigree_t *ped, swap_t *resultSwaps)
{
    double              lodThresh = INITIAL_LOD_THRESH;
    double              minThresh, maxThresh;
    linkageRelations    *relations;
    linkageRelations    *threshedRelations;
    solution_t          solution = NULL;
    int                 i;
    
    relations = buildLinkageRelations(ped, lodThresh, resultSwaps);
    
    if (phasingFromRelations(ped, relations, &solution)) {
        /* Found a solution, try lower LOD thresholds. */
        do {
            if (lodThresh <= 0.0) {
                swapPhases(ped, solution, relations, resultSwaps);
                freeSolution(solution);
                freeRelations(relations);
                return 0.0;
            }
            lodThresh -= BIG_INCREMENT;
            
            freeRelations(relations);
            relations = buildLinkageRelations(ped, lodThresh, resultSwaps);
        } while (phasingFromRelations(ped, relations, &solution));
        
        minThresh = lodThresh >= 0.0 ? lodThresh : 0.0;
        maxThresh = lodThresh + BIG_INCREMENT;
    } else {
        /* Try larger thresholds */
        do
            lodThresh += BIG_INCREMENT;
        while (!phasingFromThreshRelations(ped, relations, lodThresh,
                                           &solution));
        
        minThresh = lodThresh - BIG_INCREMENT;
        maxThresh = lodThresh;
    }
    
    /* Now refine the threshold using a binary search to find the lowest
       solvable theshold. */
    for (i = 0; i < NUM_REFINEMENTS; ++i) {
        double  mid = (maxThresh + minThresh) / 2.0;
        
        if (phasingFromThreshRelations(ped, relations, mid, &solution))
            maxThresh = mid;
        else
            minThresh = mid;
    }
    
    swapPhases(ped, solution, relations, resultSwaps);
    threshedRelations = threshRelations(relations, maxThresh);
    finalPhasing(threshedRelations, ped, resultSwaps);
    
    freeSolution(solution);
    delete threshedRelations->relations;
    delete threshedRelations;
    freeRelations(relations);
    return maxThresh;
}

/* Attempts to find the optimal phases. Uses the relations structure as a
   starting point, but first thresholds them with thresh.
 */
bool phasingFromThreshRelations(pedigree_t *ped,
                                const linkageRelations *relations,
                                double thresh, solution_t *pSwaps)
{
    linkageRelations    *rels = threshRelations(relations, thresh);
    bool                success = phasingFromRelations(ped, rels, pSwaps);
    
    delete rels->relations;
    delete rels;
    return success;
}

/* Attempts to find the optimal phases from the given set of relations.
 */
bool phasingFromRelations(pedigree_t *ped, const linkageRelations *relations,
                          solution_t *pSwaps)
{
    int                 *nodeOrder;
    solution_t          swaps = NULL;
    bool                success = false;
    
    nodeOrder = findNodeOrder(relations);
    if (nodeOrder) {
        swaps = maxLlhdPhases(relations, nodeOrder, MAX_NUM_FRONTIERS);
        if (swaps) {
            if (*pSwaps)
                freeSolution(*pSwaps);
            *pSwaps = swaps;
            success = true;
        }
    }
    
    delete [] nodeOrder;
    return success;
}

/* Builds a list of all relations in the given data set which exceed the
   threshold.
 */
linkageRelations *buildLinkageRelations(pedigree_t *ped, double lodThresh,
                                        swap_t *resultSwaps)
{
    linkageRelations    *relations = new linkageRelations;
    
    relations->numNodes = 0;
    relations->markerIds = new markerId[ped->numMarks];
    
    /* Is the following necessary? */
    for (int i = 0; i < ped->numMarks; ++i) {
        relations->markerIds[i].fatherNode = relations->markerIds[i].motherNode
            = -1;
    }
    
    for (int mark = 0; mark < ped->numMarks; ++mark) {
        bool    fatherPhase = false, motherPhase = false;
        
        for (int j = 0; j < ped->numInds; ++j) {
            obs_t   gen = ped->markers[mark].obs[j];
            
            if ((gen >> 2) != (gen & Obs0011))
                /* depends on father's phase */
                fatherPhase = true;
            if ((gen & Obs1010) >> 1 != (gen & Obs0101))
                /* depends on mother's phase */
                motherPhase = true;
        }
        
        if (fatherPhase)
            relations->markerIds[mark].fatherNode = relations->numNodes++;
        else
            relations->markerIds[mark].fatherNode = -1;
        
        if (motherPhase)
            relations->markerIds[mark].motherNode = relations->numNodes++;
        else
            relations->markerIds[mark].motherNode = -1;
        
        if (resultSwaps) {
            resultSwaps[mark] = (motherPhase ? SWAP_MOTHER_SEGREGATING : 0)
                                | (fatherPhase ? SWAP_FATHER_SEGREGATING : 0);
        }
    }
    
    relations->numRelations = 0;
    relations->relations = new relation[relations->numNodes
                                          * relations->numNodes];
    
    for (int i = 0; i < ped->numMarks; ++i) {
        for (int j = i + 1; j < ped->numMarks; ++j) {
            bool    motherPhase = relations->markerIds[i].motherNode != -1
                                    && relations->markerIds[j].motherNode != -1;
            bool    fatherPhase = relations->markerIds[i].fatherNode != -1
                                    && relations->markerIds[j].fatherNode != -1;
            
            if (motherPhase || fatherPhase) {
                if (buildRelation(ped, i, j, motherPhase, fatherPhase,
                                  relations->relations+relations->numRelations,
                                  relations->markerIds, lodThresh))
                    ++relations->numRelations;
            }
        }
    }
    
    shiftRelations(relations);
    
    return relations;
}

/* Fills in a single relation structure using the pair of markers m1 and m2. */
bool buildRelation(pedigree_t *ped, int m1, int m2, bool motherPhase,
                   bool fatherPhase, relation *rel, const markerId *markIds,
                   double lodThresh)
{
    if (motherPhase && fatherPhase) {
        double  lod, lodTwist;
        double  theta = compute2PtUnphased(ped->markers + m1, ped->markers + m2,
                                           ped->numInds, 0, &lod);
        double  thetaTwist = compute2PtUnphased(ped->markers + m1,
                                                ped->markers + m2,
                                                ped->numInds, 1, &lodTwist);
        
        if (lod <= lodThresh && lodTwist <= lodThresh)
            return false;
        
        rel->numNodePairs = 2;
        rel->nodeNdxs[0] = markIds[m1].fatherNode;
        rel->nodeNdxs[1] = markIds[m2].fatherNode;
        rel->nodeNdxs[2] = markIds[m1].motherNode;
        rel->nodeNdxs[3] = markIds[m2].motherNode;
        
        if (theta <= 0.5) {
            rel->lods[0] = lod;
            rel->lods[3] = 0.0;
        } else {
            rel->lods[3] = lod;
            rel->lods[0] = 0.0;
        }
        
        if (thetaTwist <= 0.5) {
            rel->lods[1] = lodTwist;
            rel->lods[2] = 0.0;
        } else {
            rel->lods[2] = lodTwist;
            rel->lods[1] = 0.0;
        }
        
        return true;
    } else {
        double  lod;
        double  theta = compute2PtUnphased(ped->markers + m1, ped->markers + m2,
                                           ped->numInds, 0, &lod);
        
        if (lod <= lodThresh)
            return false;
        
        rel->numNodePairs = 1;
        
        if (motherPhase) {
            rel->nodeNdxs[0] = markIds[m1].motherNode;
            rel->nodeNdxs[1] = markIds[m2].motherNode;
        } else {
            rel->nodeNdxs[0] = markIds[m1].fatherNode;
            rel->nodeNdxs[1] = markIds[m2].fatherNode;
        }
        
        if (theta <= 0.5) {
            rel->lods[0] = lod;
            rel->lods[1] = 0.0;
        } else {
            rel->lods[1] = lod;
            rel->lods[0] = 0.0;
        }
        
        return true;
    }
}

/* This shifts the values of all relations so that the maximum LOD score is
   always 0.0. This eliminates some book-keeping in maxLlhdPhases. It also fills
   in the magnitude field of the structures.
 */
void shiftRelations(linkageRelations *relations)
{
    for (int i = 0; i < relations->numRelations; ++i) {
        relation    *rel = relations->relations + i;
        int         numLods = 1 << rel->numNodePairs;
        double      maxLod = 0.0;
        
        for (int j = 0; j < numLods; ++j) {
            if (rel->lods[j] > maxLod)
                maxLod = rel->lods[j];
        }
        
        rel->magnitude = maxLod;
        for (int j = 0; j < numLods; ++j)
            rel->lods[j] -= maxLod;
    }
}

/* Creates a new linkageRelations structure which only contains those relations
   with a magnitude greater than thresh. It does not duplicate the markerIds
   field, so the src and destination structures cannot both have freeRelations
   called on them or markerIds would get delete[]'ed twice.
 */
linkageRelations *threshRelations(const linkageRelations *src, double thresh)
{
    linkageRelations    *dst = new linkageRelations;
    int                 i, j;
    
    dst->numNodes = src->numNodes;
    dst->markerIds = src->markerIds;
    dst->relations = new relation[src->numRelations];
    
    for (i = 0, j = 0; i < src->numRelations; ++i) {
        if (src->relations[i].magnitude > thresh)
            dst->relations[j++] = src->relations[i];
    }
    dst->numRelations = j;
    
    return dst;
}

void freeRelations(linkageRelations *relations)
{
    delete [] relations->markerIds;
    delete [] relations->relations;
    delete relations;
}

/* Divides the nodes into linkage groups using the relations in rels. Two nodes
   are in the same group if they appear in the same relation. The return value
   is an array of non-negative integers less than rels->numNodes such that
   nodes in the same group will have the same value in the array.
 */
int *findGroups(const linkageRelations *rels)
{
    int     *groups = new int[rels->numNodes];
    
    for (int i = 0; i < rels->numNodes; ++i)
        groups[i] = i;
    
    for (int i = 0; i < rels->numRelations; ++i) {
        relation    *r = rels->relations + i;
        int         grp = findRoot(groups, r->nodeNdxs[0]);
        
        for (int j = 1; j < r->numNodePairs * 2; ++j)
            groups[findRoot(groups, r->nodeNdxs[j])] = grp;
    }
    
    for (int i = 0; i < rels->numNodes; ++i)
        groups[i] = findRoot(groups, i);
    
    return groups;
}

/* Finds the connected components of the graph which has arcs between two nodes
   if they are directly related, i.e. they are paired in the nodeNdxs table, not
   just both present. It returns an array of size rels->numNodes with each entry
   a number which indicates the component for the corresponding node. The caller
   is responsible for calling delete [] on the returned pointer.
 */
int *findDirectGroups(const linkageRelations *rels)
{
    int     *groups = new int[rels->numNodes];
    int     i, j;
    
    for (i = 0; i < rels->numNodes; ++i)
        groups[i] = i;
    
    for (i = 0; i < rels->numRelations; ++i) {
        relation    *rel = rels->relations + i;
        
        for (j = 0; j < rel->numNodePairs; ++j) {
            int     ndx1 = findRoot(groups, rel->nodeNdxs[2 * j]);
            int     ndx2 = findRoot(groups, rel->nodeNdxs[2 * j + 1]);
            
            if (ndx1 != ndx2)
                groups[ndx1] = ndx2;
        }
    }
    
    for (i = 0; i < rels->numNodes; ++i)
        groups[i] = findRoot(groups, i);
    
    return groups;
}

int findRoot(int *groups, int ndx)
{
    while (groups[ndx] != ndx)
        ndx = groups[ndx];
    return ndx;
}

/* Chooses the order in which to traverse the nodes in order to find the
   maximum-likelihood solution.
 */
int *findNodeOrder(const linkageRelations *rels)
{
    int     numNodes = rels->numNodes;
    int     **bestOrder = new int *[numNodes];
    int     *bestMaxFrontier = new int[numNodes];
    int     *relsNdxCounts;
    int     **relsNdx = computeRelsNdx(rels, &relsNdxCounts);
    int     *groups = findGroups(rels);
    int     *numNodesGrp = new int[numNodes];
    double  *nodeTotalMagnitudes = computeNodeTotalMagnitudes(rels);
    int     *order;
    int     dst = 0;
    
    for (int i = 0; i < numNodes; ++i) {
        bestOrder[i] = NULL;
        bestMaxFrontier[i] = numNodes + 1;
        numNodesGrp[i] = 0;
    }
    
    for (int i = 0; i < numNodes; ++i) {
        int     maxFrontier;
        int     group = groups[i];
        int     n;
        int     *order;
        
        order = findNodeOrderFrom(nodeTotalMagnitudes, relsNdx, relsNdxCounts,
                                  i, &maxFrontier, &n, rels);
        
        if (maxFrontier < bestMaxFrontier[group]) {
            bestMaxFrontier[group] = maxFrontier;
            delete [] bestOrder[group];
            bestOrder[group] = order;
            numNodesGrp[group] = n;
        } else
            delete [] order;
    }
    
    delete [] groups;
    delete [] nodeTotalMagnitudes;
    for (int i = 0; i < numNodes; ++i)
        delete [] relsNdx[i];
    delete [] relsNdx;
    delete [] relsNdxCounts;
    order = new int[numNodes];
    
    for (int i = 0; i < numNodes; ++i) {
        for (int j = 0; j < numNodesGrp[i]; ++j)
            order[dst++] = bestOrder[i][j];
        delete [] bestOrder[i];
    }
    
    if (dst != numNodes) {
        fprintf(stderr, "Internal error: result in findNodeOrder is ");
        fprintf(stderr, "%d long, but should be %d\n", dst, numNodes);
    }
    
    delete [] bestMaxFrontier;
    delete [] bestOrder;
    delete [] numNodesGrp;
    
    return order;
}

/* The total magnitude for a given node is the sum of the magnitudes of all
   relations which reference that node.
   
   Caller is responsible for deleting returned pointer. */
double *computeNodeTotalMagnitudes(const linkageRelations *rels)
{
    double  *magnitudes = new double[rels->numNodes];
    
    for (int i = 0; i < rels->numNodes; ++i)
        magnitudes[i] = 0.0;
    
    for (int i = 0; i < rels->numRelations; ++i) {
        relation    *r = rels->relations + i;
        
        for (int j = 0; j < 2 * r->numNodePairs; ++j)
            magnitudes[r->nodeNdxs[j]] += r->magnitude;
    }
    
    return magnitudes;
}

/* Computes an index which contains, for each node, a list of relations in rels
   which reference that node. */
int **computeRelsNdx(const linkageRelations *rels, int **pRelsNdxCounts)
{
    int     *relsNdxCounts = new int[rels->numNodes];
    int     **relsNdx = new int *[rels->numNodes];
    int     i, j;
    
    for (i = 0; i < rels->numNodes; ++i)
        relsNdxCounts[i] = 0;
    
    for (i = 0; i < rels->numRelations; ++i) {
        for (j = 0; j < 2 * rels->relations[i].numNodePairs; ++j)
            relsNdxCounts[rels->relations[i].nodeNdxs[j]]++;
    }
    
    for (i = 0; i < rels->numNodes; ++i) {
        relsNdx[i] = new int[relsNdxCounts[i]];
        relsNdxCounts[i] = 0;
    }
    
    for (i = 0; i < rels->numRelations; ++i) {
        relation    *r = rels->relations + i;
        
        for (j = 0; j < 2 * r->numNodePairs; ++j) {
            int     node = r->nodeNdxs[j];
            
            relsNdx[node][relsNdxCounts[node]++] = i;
        }
    }
    
    *pRelsNdxCounts = relsNdxCounts;
    return relsNdx;
}

/* Finds a node order starting from the node start. The order is returned and
   the values *maxFrontier and *numNodesGrp are set to the maximum number of
   nodes in the frontier at any one time and the total number of nodes in the
   group, respectively. */
int *findNodeOrderFrom(const double *nodeTotalMagnitudes,
                       const int *const *relsNdx, const int *relsNdxCounts,
                       int start, int *maxFrontier, int *numNodesGrp,
                       const linkageRelations *rels)
{
    int     numNodes = rels->numNodes;
    int     *nodeOrder = new int[numNodes];
    bool    *used = new bool[numNodes];
    int     i;
        /* numFrontierCrosses[i] is the number of nodes j such that i and j have
           at least one linkage relation between them and used[j] != used[i]. */
    int     *numFrontierCrosses = new int[numNodes];
        /* numFrontier is the number of nodes i such that
           (used[i] && numFrontierCrosses[i] > 0). */
    int     numFrontier = 0;
        /* nodeMagnitudes[i] is the sum of all relations all of whose nodes have
           been used with the exception of i. */
    double  *nodeMagnitudes = new double[numNodes];
        /* numUnknowns[i] is the number of nodes j which are used by the
           relation i and such that !used[j]. */
    int     *numUnknowns = new int[rels->numRelations];
    int     node;
    
    for (i = 0; i < numNodes; ++i) {
        used[i] = false;
        numFrontierCrosses[i] = 0;
        nodeMagnitudes[i] = 0.0;
    }
    
    for (i = 0; i < rels->numRelations; ++i)
        numUnknowns[i] = 2 * rels->relations[i].numNodePairs;
    
    *maxFrontier = 0;
    
    node = start;
    
    for (i = 0; i < numNodes; ++i) {
        nodeOrder[i] = node;
        used[node] = true;
        updateNumFrontierCrosses(used, numFrontierCrosses, &numFrontier,
                                 relsNdx, relsNdxCounts, rels, node);
        updateNodeMagnitudes(numUnknowns, nodeMagnitudes, rels, relsNdxCounts,
                             relsNdx, node);
        
        if (numFrontier > *maxFrontier)
            *maxFrontier = numFrontier;
        
        if (numFrontier == 0) {
            ++i;
            break;
        }
        
        node = pickNextNode(rels, used, numFrontierCrosses, nodeMagnitudes,
                            nodeTotalMagnitudes);
    }
    
    delete [] used;
    delete [] numFrontierCrosses;
    delete [] nodeMagnitudes;
    delete [] numUnknowns;
    *numNodesGrp = i;
    return nodeOrder;
}

/* This picks the next node from those which haven't already been used. It
   chooses the node which has the highest number of numFrontierCrosses, with
   ties broken by nodeMagnitudes, and ties in that broken by nodeMagnitudes2
 */
int pickNextNode(const linkageRelations *rels, const bool *used,
                 const int *numFrontierCrosses, const double *nodeMagnitudes,
                 const double *nodeMagnitudes2)
{
    int     maxNumCover = -1;
    double  maxMagnitude = -DBL_MAX;
    double  maxMagnitude2 = -DBL_MAX;
    int     node = -1;
    
    for (int j = 0; j < rels->numNodes; ++j) {
        /* The following is just a dictionary order comparison between
            (numFrontierCrosses[j], nodeMagnitudes[j], nodeMagnitudes2[j])
            and (maxNumCover, maxMagnitude, maxMagnitude2). */
        if (!used[j]
              && (numFrontierCrosses[j] > maxNumCover
                    || (numFrontierCrosses[j] == maxNumCover
                          && (nodeMagnitudes[j] > maxMagnitude
                                || (nodeMagnitudes[j] == maxMagnitude
                                      && nodeMagnitudes2[j] > maxMagnitude2)))))
        {
            maxMagnitude = nodeMagnitudes[j];
            maxNumCover = numFrontierCrosses[j];
            maxMagnitude2 = nodeMagnitudes2[j];
            node = j;
        }
    }
    
    /* Error checking */
    if (node < 0 || node >= rels->numNodes)
        fprintf(stderr, "Internal error: phasing: Picked a bad node: %d\n",
                 node);
    
    return node;
}

/* Updates the numFrontierCrosses data structure to reflect having picked the
   node node as the next node. Also updates *pNumFrontier, which is the number
   of nodes which are used, but have numFrontierCrosses greater than 0.
 */
void updateNumFrontierCrosses(const bool *used, int *numFrontierCrosses,
                              int *pNumFrontier, const int *const *relsNdx,
                              const int *relsNdxCounts,
                              const linkageRelations *rels, int node)
{
    numFrontierCrosses[node] = 0;
    
    for (int i = 0; i < relsNdxCounts[node]; ++i) {
        int     relNdx = relsNdx[node][i];
        
        for (int j = 0; j < rels->relations[relNdx].numNodePairs * 2; ++j) {
            int     k = rels->relations[relNdx].nodeNdxs[j];
            
            if (k != node) {
                if (used[k]) {
                    if (--numFrontierCrosses[k] == 0)
                        --*pNumFrontier;
                } else {
                    ++numFrontierCrosses[k];
                    if (++numFrontierCrosses[node] == 1)
                        ++*pNumFrontier;
                }
            }
        }
    }
}
    
/* Updates the nodeMagnitudes array, which, for each node, is the sum of all
   relations whose only unknown is that node.
 */
void updateNodeMagnitudes(int *numUnknowns, double *nodeMagnitudes,
                          const linkageRelations *rels, const int *counts,
                          const int *const *relsNdx, int node)
{
    for (int i = 0; i < counts[node]; ++i) {
        int     relNdx = relsNdx[node][i];
        
        if (--numUnknowns[relNdx] == 1) {
            relation    *rel = rels->relations + relNdx;
            
            for (int j = 0; j < rel->numNodePairs * 2; ++j)
                nodeMagnitudes[rel->nodeNdxs[j]] += rel->magnitude;
        }
    }
}

/* Attempts to find the maximum likelihood solution to the problem described in
   rels. It traverses the nodes in the order described by nodeOrder. If it fails
   to find a solution, either because it was interrupted or because it couldn't,
   it returns NULL.
 */
solution_t maxLlhdPhases(const linkageRelations *rels, const int *nodeOrder,
                         int maxNumFrontier)
{
    partialsState_t st;
    bool            *used = new bool[rels->numNodes];
    bool            *applied = new bool[rels->numRelations];
    int             *directGroups = findDirectGroups(rels);
    solution_t      solution = NULL;
    double          collapseThresh = -DBL_MAX;
    int             i;
    
    st.partials = new partial_t[1];
    st.partials[0].lod = 0.0;
    st.partials[0].solution = newSolution(rels->numNodes);
    
    st.frontierNodes = new int[rels->numNodes];
    st.numFrontiers = 0;
    st.numNodes = rels->numNodes;
    
    for (i = 0; i < rels->numNodes; ++i)
        used[i] = false;
    for (i = 0; i < rels->numRelations; ++i)
        applied[i] = false;
    
    for (i = 0; i < rels->numNodes; ++i) {
        bool    *canBeCollapsed;
        int     node = nodeOrder[i];
        
        used[node] = true;
        if (nodeIsArbitrary(directGroups, rels->numNodes, used, node)) {
            /* We should just pick a phase for the node, instead of
               pretending like it actually matters. */
            for (int i = 0; i < (1 << st.numFrontiers); ++i)
                setSolution(st.partials[i].solution, node, false);
        } else {
            doublePartials(&st, node);
            st.frontierNodes[st.numFrontiers++] = node;
            
            /* Find the relations which can be applied and apply them. */
            for (int j = 0; j < rels->numRelations; ++j) {
                relation    *rel = rels->relations + j;
                int         k;
                
                if (applied[j])
                    continue;
                
                for (k = 0; k < 2 * rel->numNodePairs; ++k) {
                    if (!used[rel->nodeNdxs[k]])
                        break;
                }
                if (k >= 2 * rel->numNodePairs) {
                    applyRelation(rel, &st);
                    applied[j] = true;
                }
            }
            
            /* Find the nodes which can be safely collapsed because no further
               relations apply to them. */
            canBeCollapsed = new bool[st.numNodes];
            for (int j = 0; j < rels->numNodes; ++j)
                canBeCollapsed[j] = true;
            
            for (int j = 0; j < rels->numRelations; ++j) {
                if (applied[j])
                    continue;
                
                for (int k = 0; k < 2 * rels->relations[j].numNodePairs; ++k)
                    canBeCollapsed[rels->relations[j].nodeNdxs[k]] = false;
            }
            
            for (int j = 0; j < st.numFrontiers; ++j) {
                if (canBeCollapsed[st.frontierNodes[j]])
                    collapseNode(&st, j--);
            }
            
            delete [] canBeCollapsed;
            
            prematureCollapse(&st, &collapseThresh,
                              st.numFrontiers > maxNumFrontier);
            
            if (st.numFrontiers == 0) {
                if (st.partials[0].lod < collapseThresh) {
                    freeSolution(st.partials[0].solution);
                    break;
                }
                
                if (isBetweenGroups(used, applied, rels))
                    collapseThresh = -DBL_MAX;
            }
        }
    }
    
    if (i == rels->numNodes) {
        /* If we get here, it means we have found an optimal solution. */
        if (st.numFrontiers != 0) {
            fprintf(stderr, "Internal error: Wrong ");
            fprintf(stderr, "number of frontier nodes in finding phases.\n");
        }
        
        solution = st.partials[0].solution;
    }
    
    delete [] directGroups;
    delete [] used;
    delete [] applied;
    delete [] st.frontierNodes;
    delete [] st.partials;
    return solution;
}

/* Returns true iff the node has no relationships with any node which has
   already been used i.e., iff the node's phase can be assigned
   arbitrarily.
 */
bool nodeIsArbitrary(const int *groups, int numNodes, const bool *used,
                     int node)
{
    for (int i = 0; i < numNodes; ++i) {
        if (groups[i] == groups[node] && used[i] && i != node)
            return false;
    }
    
    return true;
}

bool isBetweenGroups(const bool *used, const bool *applied,
                     const linkageRelations *rels)
{
    for (int i = 0; i < rels->numRelations; ++i) {
        if (applied[i])
            continue;
        
        for (int j = 0; j < 2 * rels->relations[i].numNodePairs; ++j) {
            if (used[rels->relations[i].nodeNdxs[j]])
                return false;
        }
    }
    return true;
}

/* Adds a new node to the frontier. It doubles teh size of st.partials, where
   each entry becomes two which differ only in that they have different phases
   for the node nodeNdx.
 */
void doublePartials(partialsState_t *st, int nodeNdx)
{
    int         numPartials = 1 << st->numFrontiers;
    partial_t   *newPartials = new partial_t[2 * numPartials];
    
    for (int i = 0; i < numPartials; ++i) {
        newPartials[i] = st->partials[i];
        setSolution(newPartials[i].solution, nodeNdx, false);
        newPartials[i+numPartials].lod = st->partials[i].lod;
        newPartials[i+numPartials].solution
            = copySolution(st->partials[i].solution, st->numNodes);
        setSolution(newPartials[i+numPartials].solution, nodeNdx, true);
    }
    delete [] st->partials;
    st->partials = newPartials;
}

void applyRelation(const relation *rel, partialsState_t *st)
{
    int     numPartials = 1 << st->numFrontiers;
    
    for (int i = 0; i < numPartials; ++i)
        st->partials[i].lod += relationValue(rel, st->partials[i].solution);
}

double relationValue(const relation *rel, solution_t sol)
{
    int     ndx = 0;    /* index into the rel->lods table */
    
    for (int j = 0; j < rel->numNodePairs; ++j) {
        bool    val1 = testSolution(sol, rel->nodeNdxs[2*j]);
        bool    val2 = testSolution(sol, rel->nodeNdxs[2*j+1]);
        
        ndx *= 2;
        if ((val1 && !val2) || (!val1 && val2))
            ++ndx;
    }
    
    return rel->lods[ndx];
}

/* Removes a given node from the set of frontier nodes. It reduces the size of
   st.partials by a factor of two. For each pair of states which differ by
   only the node node, only the one with a higher LOD score is kept. */
void collapseNode(partialsState_t *st, int node)
{
    partial_t   *newPartials = new partial_t[1 << (st->numFrontiers - 1)];
    int         nodeMask = 1 << node;
    int         upperRange = 1 << (st->numFrontiers - node - 1);
    
    for (int i = 0; i < upperRange; ++i) {
        for (int j = 0; j < nodeMask; ++j) {
            int     firstNdx = i * nodeMask * 2 + j;
            int     secondNdx = firstNdx + nodeMask;
            int     newNdx = i * nodeMask + j;
            
            if (st->partials[firstNdx].lod >= st->partials[secondNdx].lod) {
                newPartials[newNdx] = st->partials[firstNdx];
                freeSolution(st->partials[secondNdx].solution);
            } else {
                newPartials[newNdx] = st->partials[secondNdx];
                freeSolution(st->partials[firstNdx].solution);
            }
        }
    }
    
    delete [] st->partials;
    st->partials = newPartials;
    
    memmove(st->frontierNodes + node, st->frontierNodes + node + 1,
            (--st->numFrontiers - node) * sizeof(int));
}

/* Executes collapseNode on nodes whose value may still influence the LOD
   score. It will do this on any node whose collapse will only throw away
   solutions with LOD less than *pThresh. Since all the relationships have a
   maximum LOD score of zero, the final LOD score of these solutions is also
   bounded by *pThresh. Furthermore, if force is true, it will be guaranteed
   to collapse at least one node, and will update *pThresh accordingly.
 */
void prematureCollapse(partialsState_t *st, double *pThresh, bool force)
{
    int     bestNode = -1;
    double  bestMaxMinLod = DBL_MAX;
    
    for (int i = 0; i < st->numFrontiers; ++i) {
        int     lowerRange = 1 << i;
        int     upperRange = 1 << (st->numFrontiers - i - 1);
        double  maxMinLod = -DBL_MAX;
        
        for (int j = 0; j < upperRange; ++j) {
            for (int k = 0; k < lowerRange; ++k) {
                int     firstNdx = j * lowerRange * 2 + k;
                int     secondNdx = firstNdx + lowerRange;
                double  lod1 = st->partials[firstNdx].lod;
                double  lod2 = st->partials[secondNdx].lod;
                double  minLod;
                
                if (lod1 > lod2)
                    minLod = lod2;
                else
                    minLod = lod1;
                
                if (minLod > maxMinLod)
                    maxMinLod = minLod;
            }
        }
        
        if (maxMinLod < *pThresh) {
            collapseNode(st, i);
            --i;
            force = false;
        } else if (maxMinLod < bestMaxMinLod) {
            bestMaxMinLod = maxMinLod;
            bestNode = i;
        }
    }
    
    if (force) {
        collapseNode(st, bestNode);
        *pThresh = bestMaxMinLod;
    }
}

/* Performs a final step of phasing on the markers. Pairs of phases which are
   not directly linked by some chain of relations have aribtrary relative
   phases after maxLlhdPhases. This function uses the 2-point statistics from
   all pairs of markers to phase these groups relative to each other. These
   pairs are used to fix the relative phases of the groups starting from the
   highest lod to the lowest.
*/
void finalPhasing(const linkageRelations *rels, pedigree_t *ped,
                  swap_t *resultSwaps)
{
    int         *groups = findDirectGroups(rels);
    relPhase_t  *relPhases;
    link_t      *links;
    int         *mapping;
    solution_t  swaps;
    int         count = 0, numLinks;
    int         i, j;
    
    /* Renumber groups to be consecutive numbers beginning with 0 */
    
    mapping = new int[rels->numNodes];
    
    for (i = 0; i < rels->numNodes; ++i) {
        if (groups[i] == i)
            mapping[i] = count++;
    }
    
    for (i = 0; i < rels->numNodes; ++i)
        groups[i] = mapping[groups[i]];
    
    delete [] mapping;
    
    /* Find the links which determine the relative phases between groups */
    
    numLinks = (count - 1) * count / 2;
    links = new link_t[numLinks];
    
    for (i = 0; i < count; ++i) {
        for (j = 0; j < i; ++j) {
            link_t  *link = links + i * (i - 1) / 2 + j;
            
            link->ndx1 = i;
            link->ndx2 = j;
            link->strength = 0.0;
            link->parity = false;
        }
    }
    
    for (i = 0; i < ped->numMarks; ++i) {
        int     nodeM1 = rels->markerIds[i].motherNode;
        int     nodeF1 = rels->markerIds[i].fatherNode;
        
        for (j = 0; j < i; ++j) {
            int     nodeM2 = rels->markerIds[j].motherNode;
            int     nodeF2 = rels->markerIds[j].fatherNode;
            bool    motherPhase = nodeM1 != -1 && nodeM2 != -1;
            bool    fatherPhase = nodeF1 != -1 && nodeF2 != -1;
            
            if (fatherPhase || motherPhase) {
                double  lod, theta;
                bool    motherTwist = false;
                
                theta = compute2PtUnphased(ped->markers + i,
                                           ped->markers + j, ped->numInds,
                                           0, &lod);
                if (fatherPhase && motherPhase) {
                    double  lodTwist, thetaTwist;
                    
                    thetaTwist = compute2PtUnphased(ped->markers + i,
                                                    ped->markers + j,
                                                    ped->numInds,
                                                    1, &lodTwist);
                    if (lodTwist > lod) {
                        motherTwist = true;
                        theta = thetaTwist;
                    }
                }
                if (motherPhase)
                    linkGroups(links, count, groups[nodeM1], groups[nodeM2],
                               lod, (theta > 0.5) ^ motherTwist);
                if (fatherPhase)
                    linkGroups(links, count, groups[nodeF1], groups[nodeF2],
                               lod, theta > 0.5);
            }
        }
    }
    
    /* use the sorted links to determine the relative phases between groups. */
    
    qsort(links, numLinks, sizeof(link_t), compareLinks);
    relPhases = new relPhase_t[count];
    
    for (i = 0; i < count; ++i) {
        relPhases[i].group = i;
        relPhases[i].parity = false;
    }
    
    for (i = 0; i < numLinks; ++i) {
        bool    parity = links[i].parity;
        int     group1 = findRootRelPhase(relPhases, links[i].ndx1, &parity);
        int     group2 = findRootRelPhase(relPhases, links[i].ndx2, &parity);
        
        if (group1 != group2) {
            relPhases[group2].group = group1;
            relPhases[group2].parity = parity;
        }
    }
    
    delete [] links;
    swaps = newSolution(rels->numNodes);
    
    for (i = 0; i < rels->numNodes; ++i) {
        bool    parity = false;
        
        findRootRelPhase(relPhases, groups[i], &parity);
        setSolution(swaps, i, parity);
    }
    
    swapPhases(ped, swaps, rels, resultSwaps);
    freeSolution(swaps);
    delete [] groups;
    delete [] relPhases;
}

void linkGroups(link_t *links, int count, int i, int j, double strength,
                bool parity)
{
    if (i != j) {
        link_t  *link;
        
        if (i < j) {
            int tmp = i;
            i = j;
            j = tmp;
        }
        
        link = links + (i - 1) * i / 2 + j;
        
        if (strength > link->strength) {
            link->strength = strength;
            link->parity = parity;
        }
    }
}

int compareLinks(const void *p1, const void *p2)
{
    const link_t    *l1 = (const link_t *)p1;
    const link_t    *l2 = (const link_t *)p2;
    
    if (l1->strength > l2->strength)
        return -1;
    else if (l1->strength < l2->strength)
        return 1;
    else
        return 0;
}

/* Finds the relative phase between a given group (indicated by ndx) and the
   root group. It also finds the cumulative parity between the group and the
   root. Note that *pParity should be initialized to 0 in order to get the
   cumulative phase or to 1 to get the opposite of the cumulative phase.
 */
int findRootRelPhase(relPhase_t *relPhases, int ndx, bool *pParity)
{
    while (relPhases[ndx].group != ndx) {
        if (pParity)
            *pParity ^= relPhases[ndx].parity;
        ndx = relPhases[ndx].group;
    }
    return ndx;
}

void swapPhases(pedigree_t *ped, solution_t swaps, const linkageRelations *rels,
                swap_t *resultSwaps)
{
    for (int i = 0; i < ped->numMarks; ++i) {
        markerId    *m = rels->markerIds + i;
        marker_t    *mark = ped->markers + i;
        
        if (m->fatherNode != -1 && testSolution(swaps, m->fatherNode)) {
            for (int j = 0; j < ped->numInds; ++j) {
                obs_t   gen = mark->obs[j];
                mark->obs[j] = (gen >> 2) | ((gen & 0x3) << 2);
            }
            if (resultSwaps)
                resultSwaps[i] ^= SWAP_FATHER;
        }
        
        if (m->motherNode != -1 && testSolution(swaps, m->motherNode)) {
            for (int j = 0; j < ped->numInds; ++j) {
                obs_t   gen = mark->obs[j];
                mark->obs[j] = ((gen & 0xa) >> 1) | ((gen & 0x5) << 1);
            }
            if (resultSwaps)
                resultSwaps[i] ^= SWAP_MOTHER;
        }
    }
}

int countSwaps(solution_t swaps, int numNodes)
{
    int     cnt = 0;
    
    for (int i = 0; i < numNodes; ++i) {
        if (testSolution(swaps, i))
            ++cnt;
    }
    
    return cnt;
}

solution_t newSolution(int numNodes)
{
    return new unsigned char[(numNodes + CHAR_BIT - 1) / CHAR_BIT];
}

void freeSolution(solution_t sol)
{
    delete [] sol;
}

solution_t copySolution(solution_t sol, int numNodes)
{
    int         numBytes = (numNodes + CHAR_BIT - 1) / CHAR_BIT;
    solution_t result = new unsigned char[numBytes];
    memcpy(result, sol, numBytes);
    return result;
}

void setSolution(solution_t sol, int pos, bool val)
{
    unsigned char   mask = 1 << (pos % CHAR_BIT);
    unsigned char   t = sol[pos / CHAR_BIT] & ~mask;
    
    sol[pos / CHAR_BIT] = val ? (t | mask) : t;
}

bool testSolution(solution_t sol, int pos)
{
    return (1 << (pos % CHAR_BIT)) & (sol[pos / CHAR_BIT]);
}
