/*
 * Copyright 2005-2006 Myriad Genetics, Inc.
 * Copyright 2010 Dustin Cartwright
 *
 * All rights reserved.
 */

package com.myriad.tmap.improve;

import java.util.Arrays;
import com.myriad.tmap.compute.*;

/** Maintains a cache of map distances for various sets of markers. For each
 * list of marker names and whether or not its split, it maintains a Map object.
 * The order in the cache is always maintained as most recently used to least
 * recently used. The built-in lock is used to maintain consistency of the
 * cache. */
class MapCache {
    MapCache()
    {
        Map             emptyMap = new Map(new MarkerPosition[0]);
        CacheElement    emptyCacheElem;
        
        emptyCacheElem = new CacheElement(new String[0], false, emptyMap);
        cache = new CacheElement[kCacheSize];
        for (int i = 0; i < cache.length; ++i)
            cache[i] = emptyCacheElem;
    }

    /** Looks for a map in the cache. Returns the map or null if the map can't
     * be found in cache.
     * @param markerNames array of markers for map
     * @param split whether or not map build with split distances.
     */
    synchronized Map getMap(String[] markerNames, boolean split)
    {
        for (int i = 0; i < cache.length; ++i) {
            if (cache[i].split == split
                    && Arrays.equals(cache[i].markerNames, markerNames)) {
                Map     map = cache[i].result;
                /* Push result on beginning of cache, replacing previous
                 * entry */
                cacheMap(markerNames, split, map, i);
                return map;
            }
        }

        return null;
    }

    /** Adds a map to the cache.
        @param markerNames the array of markers which make up the map
        @param split whether or not the map was built with separate distances
        for the mother and the father.
        @param map the resulting map.
     */
    synchronized void cacheMap(String[] markerNames, boolean split,
            Map map)
    {
        cacheMap(markerNames, split, map, cache.length - 1);
    }

    /** Adds the specified pairing of markerNames and map to the cache,
        replacing the cache element at the repNdx.
        @param markerNames the array of markers which make up the map
        @param split whether or not the map was built with separate distances
        for the mother and the father.
        @param map the resulting map.
        @param repNdx the index of the cache element to remove to make room for
        the current cache element.
     */
    private void cacheMap(String[] markerNames, boolean split, Map map,
                          int repNdx)
    {
        for (int i = repNdx; i > 0; --i)
            cache[i] = cache[i - 1];
        cache[0] = new CacheElement(markerNames, split, map);
    }

    /** Represents a single element in the cache, i.e. a mapping from a list of
        marker names and a split flag to the resulting map.
     */
    private class CacheElement {
        CacheElement(String[] aMarkerNames, boolean aSplit, Map aResult)
        {
            markerNames = aMarkerNames;
            split = aSplit;
            result = aResult;
        }
        
        String[]    markerNames;
        boolean     split;
        Map         result;
    }

    private CacheElement[]  cache;
    private final static int    kCacheSize = 8;
};
