/*
 * Copyright 2005-2006 Myriad Genetics, Inc.
 * Copyright 2010 Dustin Cartwright
 *
 * All rights reserved.
 */

package com.myriad.tmap.improve;

import java.awt.*;
import java.awt.event.*;
import javax.swing.*;
import java.text.NumberFormat;
import com.myriad.tmap.compute.*;
import com.myriad.tmap.Quitter;

/** A MapComponent displays a list of markers with their positions and estimated
    error rates. Markers can be deselected and the map will be recomputed
    accordingly. */
public class MapComponent extends JComponent implements MapReceivable {
    /** Creates a new map component given an already computed map.
     */
    public MapComponent(Pedigree aPed, String[] markerNames, boolean split,
            String aName)
    {
        JPanel      southPanel;
        Box         mainBox;
        JButton     reselectAllButton, improveButton, saveButton, flipButton;
        
        ped = aPed;
        pedThr = new PedigreeThread(this, ped);
        name = aName;
        
        mainPanel = new JPanel();
        
        mainBox = Box.createVerticalBox();
        mainBox.add(mainPanel);
        mainBox.add(Box.createGlue());
        
        setLayout(new BorderLayout());
        totalDistLabel = new JLabel("total distance");
        add(totalDistLabel, BorderLayout.NORTH);
        add(new JScrollPane(mainBox), BorderLayout.CENTER);
        
        southPanel = new JPanel();
        add(southPanel, BorderLayout.SOUTH);
        
        reselectAllButton = new JButton("Reselect All");
        reselectAllButton.addActionListener(new ReselectAllListener());
        southPanel.add(reselectAllButton);
        
        improveButton = new JButton("Improve");
        improveButton.addActionListener(new ImproveListener());
        southPanel.add(improveButton);
        
        saveButton = new JButton("Save");
        saveButton.addActionListener(new SaveListener());
        southPanel.add(saveButton);
        
        flipButton = new JButton("Flip");
        flipButton.addActionListener(new FlipListener());
        southPanel.add(flipButton);
        
        splitControl = new JCheckBox("Split");
        splitControl.setSelected(split);
        splitControl.addActionListener(new SplitListener());
        southPanel.add(splitControl);
        
        buildMarkerControls(markerNames);
        buildMainPanel();
        pedThr.computeMap(markerNames, split);
    }
    
    /** Creates a new frame containing the MapComponent.
        @param rel if non-null, it will position the window offset from this
        point.
     */
    public void putInNewFrame(Point rel)
    {
        JFrame  frame;
        
        frame = new JFrame("Linkage group " + name + " from "
                                + ped.getFilename());
        
        if (rel != null)
            frame.setLocation(rel.x + 50, rel.y + 50);
        
        frame.getContentPane().add(this);
        frame.setDefaultCloseOperation(JFrame.DISPOSE_ON_CLOSE);
        frame.addWindowListener(new Quitter());
        frame.setSize(450, 800);
        frame.setVisible(true);
    }
    
    /** Creates the controls used for the display and stores them in the markers
        array. They are not yet placed in the main panel, so they won't be
        displayed. */
    private void buildMarkerControls(String[] markerNames)
    {
        MarkerSwitchListener    markerSwitch = new MarkerSwitchListener();
        
        markers = new Marker2[markerNames.length];
        
        for (int i = 0; i < markerNames.length; ++i) {
            JCheckBox   checkBox = new JCheckBox(markerNames[i]);
            
            checkBox.addActionListener(markerSwitch);
            checkBox.setSelected(true);
            
            markers[i] = new Marker2(markerNames[i], checkBox);
        }
        flipped = false;
    }
    
    /** Fills the main panel with the components for all the markers and their
        positions and estimated error rates.
     */
    private void buildMainPanel()
    {
        boolean splitDists = splitControl.isSelected();
        boolean splitErrors = splitControl.isSelected();
        int     width = 3;
        
        if (splitDists) ++width;
        if (splitErrors) ++width;
        
        mainPanel.setLayout(new GridLayout(markers.length, width));
        
        for (int i = 0; i < markers.length; ++i) {
            mainPanel.add(markers[i].checkBox);
            mainPanel.add(markers[i].posMothField);
            if (splitDists)
                mainPanel.add(markers[i].posFathField);
            mainPanel.add(markers[i].errorMothField);
            if (splitErrors)
                mainPanel.add(markers[i].errorFathField);
        }
        
        Dimension   max = mainPanel.getMaximumSize();
        max.height = mainPanel.getPreferredSize().height;
        mainPanel.setMaximumSize(max);
        
        mainPanel.revalidate();
    }
    
    /** Returns the marker names of all currently selected markers
     */
    private String[] getMarkerNames()
    {
        java.util.ArrayList names = new java.util.ArrayList();
        
        for (int i = 0; i < markers.length; ++i) {
            if (markers[i].checkBox.isSelected())
                names.add(markers[i].name);
        }
        
        return (String[]) names.toArray(new String[0]);
    }
    
    /** Updates the distances and error rates in the display according to the
        values in the map parameters.
       
        The markers in map are assumed to be a subset of the markers in the
        markers element.

        Implements the MapReceivable protocol.
        @see com.myriad.tmap.improve.MapReceivable */
    public void mapResult(Map map)
    {
        int             j = 0;
        double          totalDistMoth, totalDistFath;
        NumberFormat    distFmt = NumberFormat.getInstance();
        NumberFormat    eFmt = NumberFormat.getPercentInstance();
        String          headerStr;

        /* Put the map in the same order as currently being displayed */
        if (flipped)
            map = map.reverse();
        
        if (map.markers.length > 0) {
            totalDistMoth = map.markers[map.markers.length - 1].posMoth;
            totalDistFath = map.markers[map.markers.length - 1].posFath;
        } else
            totalDistMoth = totalDistFath = 0.0;
        
        distFmt.setMaximumFractionDigits(1);
        eFmt.setMaximumFractionDigits(1);
        
        for (int i = 0; i < markers.length; ++i) {
            if (j < map.markers.length
                  && markers[i].name.equals(map.markers[j].name)) {
                MarkerPosition  mark = map.markers[j];
                
                markers[i].posMothField.setText(distFmt.format(mark.posMoth));
                markers[i].posFathField.setText(distFmt.format(mark.posFath));
                markers[i].errorMothField.setText(eFmt.format(mark.errorMoth));
                markers[i].errorFathField.setText(eFmt.format(mark.errorFath));
                ++j;
            } else {
                markers[i].posMothField.setText("--");
                markers[i].posFathField.setText("--");
                markers[i].errorMothField.setText("--");
                markers[i].errorFathField.setText("--");
            }
        }
        
        headerStr = "total distance = " + distFmt.format(totalDistMoth);
        if (totalDistMoth != totalDistFath)
            headerStr = headerStr + "/" + distFmt.format(totalDistFath);
        totalDistLabel.setText(headerStr);
    }
    
    private void recomputeMap()
    {
        String[]    names = getMarkerNames();

        if (flipped) {
            /* We always send the marker names off in the same order as when the
             * object was initialized, so that caching works well. */
            for (int i = 0, j = names.length - 1; i < j; ++i, --j) {
                String      tmp = names[i];
                names[i] = names[j];
                names[j] = tmp;
            }
        }

        pedThr.computeMap(names, splitControl.isSelected());
    }
    
    private class MarkerSwitchListener implements ActionListener {
        MarkerSwitchListener()
        {
        }
        
        public void actionPerformed(ActionEvent event)
        {
            recomputeMap();
        }
    }
    
    private class ImproveListener implements ActionListener {
        public void actionPerformed(ActionEvent event)
        {
            String[]            markerNames = getMarkerNames();
            boolean             split = splitControl.isSelected();
            MakeNewMapThread    thr;
            
            thr = new MakeNewMapThread(ped, markerNames, split,
                                       getLocationOnScreen(), true, name);
        }
    }
    
    private class ReselectAllListener implements ActionListener {
        public void actionPerformed(ActionEvent event)
        {
            for (int i = 0; i < markers.length; ++i)
                markers[i].checkBox.setSelected(true);
            recomputeMap();
        }

    }
    
    private class SaveListener implements ActionListener {
        public void actionPerformed(ActionEvent event)
        {
            JFileChooser    chooser = new JFileChooser();
            int             result;
            
            chooser.setDialogTitle("Choose save location");
            if (chooser.showSaveDialog(null) == JFileChooser.APPROVE_OPTION)
                ped.writeMap(getMarkerNames(), splitControl.isSelected(),
                             chooser.getSelectedFile());
        }
    }
    
    private class FlipListener implements ActionListener {
        public void actionPerformed(ActionEvent event)
        {
            for (int i = 0, j = markers.length - 1; i < j; ++i, --j) {
                Marker2     tmp = markers[i];
                markers[i] = markers[j];
                markers[j] = tmp;
            }
            flipped = !flipped;

            mainPanel.removeAll();
            buildMainPanel();
            recomputeMap();
        }
    }
    
    private class SplitListener implements ActionListener {
        public void actionPerformed(ActionEvent event)
        {
            mainPanel.removeAll();
            buildMainPanel();
            recomputeMap();
        }
    }
    
    /** Contains the marker names and the associated components for that marker.
     */
    private class Marker2 {
        public Marker2(String aName, JCheckBox aCheckBox)
        {
            name = aName;
            checkBox = aCheckBox;
            posMothField = new JLabel("");
            posFathField = new JLabel("");
            errorMothField = new JLabel("");
            errorFathField = new JLabel("");
        }
        
        String      name;
        JCheckBox   checkBox;
        JLabel      posMothField;
        JLabel      posFathField;
        JLabel      errorMothField;
        JLabel      errorFathField;
    }
    
    private JLabel          totalDistLabel;
    private Marker2[]       markers;
    private boolean         flipped;    // flipped relative to initial order?
    private Pedigree        ped;
    private PedigreeThread  pedThr;
    private JPanel          mainPanel;
    private JCheckBox       splitControl;
    private String          name;
}
