/*
 * Copyright 2005-2006 Myriad Genetics, Inc.
 *
 * All rights reserved.
 */

#include "ped.h"
#include <stdlib.h>
#include <string.h>

void freePed(pedigree_t *ped)
{
    int     i;
    
    for (i = 0; i < ped->numMarks; ++i)
        free(ped->markers[i].name);
    free(ped->markers[0].obs);
    free(ped->markers);
}

/* This function combines two pedigrees into a single one. The new pedigree
   contains the union of the markers in the two source pedigrees. Markers which
   only appear in one of the source pedigrees are marked as unknown for the
   individuals in the other source pedigree.
 */
void mergePeds(pedigree_t *dst, const pedigree_t *src1, const pedigree_t *src2)
{
    int     i, j, k;
    int     *used = (int *)malloc(src2->numMarks * sizeof(int));
    obs_t   *obs;
    
    dst->numInds = src1->numInds + src2->numInds;
    dst->markers = (marker_t *)malloc((src1->numMarks + src2->numMarks)
                                          * sizeof(marker_t));
    obs = (obs_t *)malloc((src1->numMarks + src2->numMarks) * dst->numInds
			    * sizeof(obs_t));
    
    for (i = 0; i < src2->numMarks; ++i)
        used[i] = FALSE;
    
    for (i = 0; i < src1->numMarks; ++i) {
        for (j = 0; j < src2->numMarks; ++j) {
            if (strcmp(src1->markers[i].name, src2->markers[j].name) == 0) {
                used[j] = TRUE;
                break;
            }
        }
        
        dst->markers[i].name = (char *)malloc(strlen(src1->markers[i].name)+1);
        strcpy(dst->markers[i].name, src1->markers[i].name);
        
	dst->markers[i].obs = obs + i * dst->numInds;
        memcpy(dst->markers[i].obs, src1->markers[i].obs,
               src1->numInds * sizeof(obs_t));
        if (j < src2->numMarks)
            memcpy(dst->markers[i].obs + src1->numInds, src2->markers[j].obs,
                   src2->numInds * sizeof(obs_t));
        else {
            int     k;
            for (k = src1->numInds; k < dst->numInds; ++k)
                dst->markers[i].obs[k] = OBSERVATION_UNKNOWN;
        }
    }
    
    for (j = 0; j < src2->numMarks; ++j) {
        if (!used[j]) {
            dst->markers[i].name
                = (char *)malloc(strlen(src2->markers[j].name) + 1);
            strcpy(dst->markers[i].name, src2->markers[j].name);
            
	    dst->markers[i].obs = obs + i * dst->numInds;
            for (k = 0; k < src1->numInds; ++k)
                dst->markers[i].obs[k] = OBSERVATION_UNKNOWN;
            memcpy(dst->markers[i].obs + src1->numInds, src2->markers[j].obs,
                   src2->numInds * sizeof(obs_t));
            ++i;
        }
    }
    
    free(used);
    dst->markers = (marker_t *)realloc(dst->markers, i * sizeof(marker_t));
    dst->numMarks = i;
}

void swapPeds(pedigree_t *p1, pedigree_t *p2)
{
    int     tmp;
    marker_t    *tmpMark;
    
    tmp = p1->numMarks;
    p1->numMarks = p2->numMarks;
    p2->numMarks = tmp;
    
    tmp = p1->numInds;
    p1->numInds = p2->numInds;
    p2->numInds = tmp;
    
    tmpMark = p1->markers;
    p1->markers = p2->markers;
    p2->markers = tmpMark;
}

/* Determines in which parents the marker is informative, and returns the
   result in *mother and *father.
 */
void markerWhereInform(marker_t *m, int numInds, int *mother, int *father)
{
    int     i;
    
    *mother = *father = FALSE;
    for (i = 0; i < numInds; ++i) {
        obs_t   obs = m->obs[i];
        if ((obs >> 2) != (obs & OBSERVATION_0011)) {
            *father = TRUE;
            if (*mother)
                return;
        }
        if ((obs & OBSERVATION_1010) >> 1 != (obs & OBSERVATION_0101)) {
            *mother = TRUE;
            if (*father)
                return;
        }
    }
}

/* Changes the marker to have the opposite maternal phase. */
void swapMaternalPhase(marker_t *mark, int numInd)
{
    int     i;
    
    for (i = 0; i < numInd; ++i) {
        obs_t   obs = mark->obs[i];
        mark->obs[i] = ((obs & 0xa) >> 1) | ((obs & 0x5) << 1);
    }
}

/* Changes the marker to have the opposite paternal phase. */
void swapPaternalPhase(marker_t *mark, int numInd)
{
    int     i;
    
    for (i = 0; i < numInd; ++i) {
        obs_t   obs = mark->obs[i];
        mark->obs[i] = (obs >> 2) | ((obs & 0x3) << 2);
    }
}
