/*
 * Copyright 2005-2006 Myriad Genetics, Inc.
 *
 * All rights reserved.
 */

#include "twoPoint.h"
#include <stdio.h>
#include <math.h>

#define INITIAL_THETA 0.2
        /* Ceases the EM-search when theta varies by less than THETA_EPS */
#define THETA_EPS 0.001

/* Computes the two point maximum likelihood theta. The parameter twists should
   be a positive integer less than 4, i.e. two bits. The least significant bit
   signifies whether the relative phase on the mother's side should be flipped,
   and the other bit does the same thing on the father's side. If pLod is
   not null, it additionally computes the lod score, and stores that in
   *pLod.
   
   See also compute2PtUnphased and compute2PtPhased, which specialize for the
   most common cases.
 */
double compute2Pt(marker_t *m1, marker_t *m2, int numInds, int twists,
                  double initialTheta, double *pLod)
{
    int             counts[16];
    int             i, j, k;
    double          theta, newTheta, thetaDiff;
    const double    propRecombs[NUM_GENOTYPES] = {0.0, 0.5, 0.5, 1.0};
    double          inertia;
    
    for (i = 0; i < 16; ++i)
        counts[i] = 0;
    
    /* Aggregates individuals in different recombination categories. */
    for (i = 0; i < numInds; ++i) {
        obs_t   gen1 = m1->obs[i];
        obs_t   gen2 = m2->obs[i];
        int     possRecomb = 0;
        
        for (j = 0; j < NUM_GENOTYPES; ++j) {
            if (!(gen1 & (1 << j)))
                continue;
            for (k = 0; k < NUM_GENOTYPES; ++k) {
                if (!(gen2 & (1 << k)))
                    continue;
                
                possRecomb |= 1 << (j ^ k ^ twists);
            }
        }
        if (possRecomb == 0) {
            printf("Warning: data set includes impossible genotype.");
        }
        
        counts[possRecomb]++;
    }
    
    /* The following 4 classes provide no information about the presence or
       absence of recombination in one of the parents. Thus, in the EM
       algorithm, the estimated number of recombinations in that gamete is
       always the old value of the recombination fraction. This effectively
       slows convergence by adding an element of "inertia." The following
       quantity is used in the EM algorithm to cancel this inertia when
       computing the new recombination fraction. */
    inertia = ((double) (counts[3] + counts[5] + counts[10] + counts[12]))/2.0;
    
    newTheta = initialTheta;
    
    /* uses EM-algorithm to estimate maximum likelihood value of theta. */
    do {
        double  expNumRecombs = 0.0;
        double  llhds[4];
        
        theta = newTheta;
        if (pLod)
            *pLod = 0.0;
        
        llhds[0] = (1.0 - theta) * (1.0 - theta);
        llhds[1] = llhds[2] = theta * (1.0 - theta);
        llhds[3] = theta * theta;
        
        /* There is no need to cover the case i = 15 which corresponds to any
           possible recombination pattern. These individuals are not
           informative, so there is no point in using them. In other words, the
           will always produce a value of a = theta and totalLlhd = llhdHalf =
           1.0, and so it just adds an element of inertia to the EM algorithm.
         */
        for (i = 1; i < 15; ++i) {
            double  totalLlhd = 0.0;
            double  a = 0.0;
            double  llhdHalf = 0.0;
            
            if (counts[i] == 0)
                continue;
            
            for (j = 0; j < NUM_GENOTYPES; ++j) {
                if (i & (1 << j)) {
                    a += propRecombs[j] * llhds[j];
                    totalLlhd += llhds[j];
                    llhdHalf += 0.25;
                }
            }
            
            expNumRecombs += counts[i] * a / totalLlhd;
            if (pLod)
                *pLod += counts[i] * log10(totalLlhd / llhdHalf);
        }
        
        newTheta = (expNumRecombs - inertia * theta)
                        / (numInds - counts[15] - inertia);
        thetaDiff = theta - newTheta;
    } while(thetaDiff > THETA_EPS || thetaDiff < -THETA_EPS);
    
    if (isnan(theta)) {
        fprintf(stderr, "Internal error: Recombination fraction set to NaN\n");
    } else {
        if (pLod && isnan(*pLod))
            fprintf(stderr, "Internal error: LOD score set to NaN\n");
        
        if (theta < 0.0) theta = 0.0;
        if (theta > 1.0) theta = 1.0;
    }
    
    return theta;
}

/* Computes the 2-point statistics of a pair of unphased markers. */
double compute2PtUnphased(marker_t *m1, marker_t *m2, int numInds,
                               int twists, double *pLod)
{
    return compute2Pt(m1, m2, numInds, twists, 0.5, pLod);
}

/* Computest the 2-point statisics for a pair of markers assuming values
   reasonable if the markers are phase-known.
 */
double compute2PtPhased(marker_t *m1, marker_t *m2, int numInds, double *pLod)
{
    return compute2Pt(m1, m2, numInds, 0, INITIAL_THETA, pLod);
}
